@@ -263,7 +263,8 @@ set_var_from_str(char *str, char **ptr, numeric *dest)
 			return -1;
 		}
 		(*ptr) = endptr;
-		if (exponent >= INT_MAX / 2 || exponent <= -(INT_MAX / 2))
+		if (exponent > NUMERIC_MAX_PRECISION ||
+			exponent < -NUMERIC_MAX_PRECISION)
 		{
 			errno = PGTYPES_NUM_BAD_NUMERIC;
 			return -1;
